@@ -258,6 +258,7 @@ struct nfs_client *nfs_alloc_client(const struct nfs_client_initdata *cl_init)
 	clp->cl_proto = cl_init->proto;
 	clp->cl_nconnect = cl_init->nconnect;
 	clp->cl_max_connect = cl_init->max_connect ? cl_init->max_connect : 1;
+	clp->cl_pconnect = cl_init->pconnect ? cl_init->pconnect : 1;
 	clp->cl_net = get_net(cl_init->net);
 	clp->cl_xprt_id = cl_init->xprt_id;
 
@@ -637,6 +638,9 @@ int nfs_create_rpc_client(struct nfs_client *clp,
 			RPC_CLNT_CREATE_NO_IDLE_TIMEOUT : 0,
 	};
 
+	if (args.version == 3)
+		args.pconnect = clp->cl_pconnect;
+
 	if (test_bit(NFS_CS_DISCRTRY, &clp->cl_flags))
 		args.flags |= RPC_CLNT_CREATE_DISCRTRY;
 	if (test_bit(NFS_CS_NO_RETRANS_TIMEOUT, &clp->cl_flags))
@@ -801,6 +805,7 @@ static int nfs_init_server(struct nfs_server *server,
 		.timeparms = &timeparms,
 		.cred = server->cred,
 		.nconnect = ctx->nfs_server.nconnect,
+		.pconnect = ctx->nfs_server.pconnect,
 		.init_flags = (1UL << NFS_CS_REUSEPORT),
 		.local_ports = ctx->local_ports,
 		.local_ports_failover = ctx->local_ports_failover,
@@ -809,7 +814,6 @@ static int nfs_init_server(struct nfs_server *server,
 		.remote_ports_offset_provided = ctx->nfs_server.remoteports_offset_provided,
 		.spread_reads = ctx->nfs_server.spread_reads,
 		.spread_writes = ctx->nfs_server.spread_writes,
-		.no_idle_xprt = ctx->no_idle_xprt,
 		.mdconnect = ctx->nfs_server.mdconnect,
 	};
 	struct nfs_client *clp;
@@ -822,17 +826,22 @@ static int nfs_init_server(struct nfs_server *server,
 	if (ctx->flags & NFS_MOUNT_NOSHARE_XPRT)
 		set_bit(NFS_CS_NO_SHARE, &cl_init.init_flags);
 
-	/* Allocate or find a client reference we can use */
-	clp = nfs_get_client(&cl_init);
-	if (IS_ERR(clp))
-		return PTR_ERR(clp);
-
 	if (cl_init.mdconnect + cl_init.nconnect > NFS_MAX_CONNECTIONS) {
 		cl_init.mdconnect = NFS_MAX_CONNECTIONS - cl_init.nconnect;
 		printk(KERN_WARNING "NFS: Clamping down mdconnect to '%d'\n",
 		       cl_init.mdconnect);
 	}
 
+	cl_init.no_idle_xprt = get_no_idle_xprt(ctx);
+
+	if (ctx->local_ports && ctx->local_ports->nr > ctx->nfs_server.nconnect)
+		printk(KERN_WARNING "NFS: nconnect value is lower than available local ports\n");
+
+	/* Allocate or find a client reference we can use */
+	clp = nfs_get_client(&cl_init);
+	if (IS_ERR(clp))
+		return PTR_ERR(clp);
+
 	server->nfs_client = clp;
 
 	/* Initialise the client representation from the mount data */
