@@ -94,6 +94,7 @@ struct nfs_client_initdata {
 	u32 minorversion;
 	unsigned int nconnect;
 	unsigned int max_connect;
+	unsigned int pconnect;
 	u32 server_idx;				/* NFSv4.1 server multipath index */
 	struct net *net;
 	const struct rpc_timeout *timeparms;
@@ -126,10 +127,10 @@ struct nfs_fs_context {
 	char			*fscache_uniq;
 	unsigned short		protofamily;
 	unsigned short		mountfamily;
-	bool                    has_sec_mnt_opts;
-	struct rpc_portgroup *local_ports;
-	struct rpc_portgroup *remote_ports;
-	bool			no_idle_xprt;
+	bool			has_sec_mnt_opts;
+	struct rpc_portgroup	*local_ports;
+	struct rpc_portgroup	*remote_ports;
+	int			no_idle_xprt;
 	bool			local_ports_failover;
 	u32			server_idx;
 	unsigned int		xprt_id;
@@ -159,6 +160,7 @@ struct nfs_fs_context {
 		unsigned short		nconnect;
 		unsigned short		mdconnect;
 		unsigned short		max_connect;
+		unsigned short		pconnect;
 		unsigned short		export_path_len;
 		u32			remoteports_offset;
 		bool			remoteports_offset_provided;
@@ -901,6 +903,18 @@ static inline bool nfs_error_is_fatal_on_server(int err)
 	return nfs_error_is_fatal(err);
 }
 
+static inline bool get_no_idle_xprt(const struct nfs_fs_context *ctx)
+{
+	if (ctx->local_ports && ctx->local_ports->nr > 1 &&
+	    ctx->nfs_server.protocol == XPRT_TRANSPORT_RDMA &&
+	    ctx->no_idle_xprt == -1)
+		return true;
+	else if (ctx->no_idle_xprt != -1)
+		return (bool) ctx->no_idle_xprt;
+	else
+		return false;
+}
+
 /*
  * Select between a default port value and a user-specified port value.
  * If a zero value is set, then autobind will be used.
