@@ -909,7 +909,7 @@ static int nfs4_set_client(struct nfs_server *server,
 		const char *ip_addr,
 		int proto, const struct rpc_timeout *timeparms,
 		u32 minorversion, unsigned int nconnect,
-		unsigned int max_connect,
+		unsigned int max_connect, unsigned int pconnect,
 		struct net *net, unsigned int xprt_id,
 		struct nfs_fs_context *ctx)
 {
@@ -926,6 +926,7 @@ static int nfs4_set_client(struct nfs_server *server,
 		.cred = server->cred,
 		.server_idx = server->idx,
 		.xprt_id = xprt_id,
+		.pconnect = pconnect,
 	};
 	struct nfs_client *clp;
 
@@ -933,11 +934,15 @@ static int nfs4_set_client(struct nfs_server *server,
 		__set_bit(NFS_CS_REUSEPORT, &cl_init.init_flags);
 	else
 		cl_init.max_connect = max_connect;
-	if (proto == XPRT_TRANSPORT_TCP)
+	switch (proto) {
+	case XPRT_TRANSPORT_RDMA:
+	case XPRT_TRANSPORT_TCP:
 		cl_init.nconnect = nconnect;
+		break;
+	}
 	if (ctx) {
 		cl_init.local_ports = ctx->local_ports;
-		cl_init.no_idle_xprt = ctx->no_idle_xprt;
+		cl_init.no_idle_xprt = get_no_idle_xprt(ctx);
 		cl_init.local_ports_failover = ctx->local_ports_failover;
 		cl_init.mdconnect = 0; /* Not relevant for NFSv4.x */
 	}
@@ -1186,6 +1191,7 @@ static int nfs4_init_server(struct nfs_server *server, struct fs_context *fc)
 				ctx->minorversion,
 				ctx->nfs_server.nconnect,
 				ctx->nfs_server.max_connect,
+			 	ctx->nfs_server.pconnect,
 				fc->net_ns, ctx->xprt_id, ctx);
 	if (error < 0)
 		return error;
@@ -1269,8 +1275,38 @@ struct nfs_server *nfs4_create_multi_server(struct fs_context *fc)
 		goto out;
 	}
 	port = rpc_get_port(&ctx->nfs_server.address);
-	reminder = orig_nconnect % remoteports->nr;
-	divised_per_client = orig_nconnect / remoteports->nr;
+
+	if (ctx->nfs_server.pconnect > 0) {
+		uint32_t needed_clients;
+		/*
+		 * `pconnect` passed - it specifies the minimum number of connections
+		 * per remote port. We validate and calculate how many remote ports
+		 * to actually use based on nconnect/pconnect ratio.
+		 *
+		 * Examples:
+		 * - nconnect=16,pconnect=4,remoteports=4 -> use 4 ports, 4 conn each
+		 * - nconnect=8,pconnect=4,remoteports=4 -> use 2 ports, 4 conn each
+		 * - nconnect=4,pconnect=4,remoteports=4 -> use 1 port, 4 conn
+		 * - nconnect=16,pconnect=2,remoteports=4 -> use 4 ports, 4 conn each
+		 */
+
+		/* Validate: pconnect cannot exceed nconnect */
+		if (ctx->nfs_server.pconnect > orig_nconnect)
+			return ERR_PTR(-EINVAL);
+
+		needed_clients = (orig_nconnect + ctx->nfs_server.pconnect - 1)
+			/ ctx->nfs_server.pconnect;
+
+		reminder = orig_nconnect % needed_clients;
+		divised_per_client = orig_nconnect / needed_clients;
+	} else {
+		/*
+		 * `pconnect` not passed, we set nconnect per client based on
+		 * amount of remoteports and nconnect.
+		 */
+		reminder = orig_nconnect % remoteports->nr;
+		divised_per_client = orig_nconnect / remoteports->nr;
+	}
 
 	remoteport_offset = rpc_calc_portgroup_offset(
 		fc->net_ns, 0, remoteports,
@@ -1291,6 +1327,10 @@ struct nfs_server *nfs4_create_multi_server(struct fs_context *fc)
 		const int i = remoteport_offset;
 		const uint32_t client_nconnect = divised_per_client
 			+ ((idx < reminder) ? 1 : 0);
+		/*
+		 * One iteration of this loop creates a client with
+		 * `client_nconnect` connections
+		 */
 		ctx->nfs_server.nconnect = client_nconnect;
 		ctx->nfs_server.max_connect = divised_per_client + 1;
 
@@ -1377,6 +1417,7 @@ struct nfs_server *nfs4_create_referral_server(struct fs_context *fc)
 				parent_client->cl_mvops->minor_version,
 				parent_client->cl_nconnect,
 				parent_client->cl_max_connect,
+				parent_client->cl_pconnect,
 				parent_client->cl_net,
 				parent_client->cl_xprt_id, ctx);
 	if (!error)
@@ -1394,6 +1435,7 @@ struct nfs_server *nfs4_create_referral_server(struct fs_context *fc)
 				parent_client->cl_mvops->minor_version,
 				parent_client->cl_nconnect,
 				parent_client->cl_max_connect,
+				parent_client->cl_pconnect,
 				parent_client->cl_net,
 				parent_client->cl_xprt_id, ctx);
 	if (error < 0)
@@ -1492,7 +1534,8 @@ int nfs4_update_server(struct nfs_server *server, const char *hostname,
 	error = nfs4_set_client(server, hostname, sap, salen, buf,
 				clp->cl_proto, clnt->cl_timeout,
 				clp->cl_minorversion,
-				clp->cl_nconnect, clp->cl_max_connect, 
+				clp->cl_nconnect, clp->cl_max_connect,
+			 	clp->cl_pconnect,
 				net, clp->cl_xprt_id, NULL);
 	clear_bit(NFS_MIG_TSM_POSSIBLE, &server->mig_status);
 	if (error != 0) {
