@@ -9,7 +9,7 @@
 #include <linux/nfs_fs.h>
 #include <linux/nfs_page.h>
 #include <linux/module.h>
-#include <linux/sched/mm.h>
+#include <linux/sched.h>
 #include <linux/sunrpc/metrics.h>
 #include <linux/old-kernel.h>
 
@@ -26,16 +26,8 @@
 #define NFSDBG_FACILITY         NFSDBG_PNFS_LD
 
 #define FF_LAYOUT_POLL_RETRY_MAX     (15*HZ)
-#define FF_LAYOUTRETURN_MAXERR 20
 
-static void ff_layout_read_record_layoutstats_done(struct rpc_task *task,
-		struct nfs_pgio_header *hdr);
-static int ff_layout_mirror_prepare_stats(struct pnfs_layout_hdr *lo,
-			       struct nfs42_layoutstat_devinfo *devinfo,
-			       int dev_limit);
-static void ff_layout_encode_ff_layoutupdate(struct xdr_stream *xdr,
-			      const struct nfs42_layoutstat_devinfo *devinfo,
-			      struct nfs4_ff_layout_mirror *mirror);
+static struct group_info	*ff_zero_group;
 
 static struct pnfs_layout_hdr *
 ff_layout_alloc_layout_hdr(struct inode *inode, gfp_t gfp_flags)
@@ -350,9 +342,11 @@ static void ff_layout_sort_mirrors(struct nfs4_ff_layout_segment *fls)
 	for (i = 0; i < fls->mirror_array_cnt - 1; i++) {
 		for (j = i + 1; j < fls->mirror_array_cnt; j++)
 			if (fls->mirror_array[i]->efficiency <
-			    fls->mirror_array[j]->efficiency)
+			    fls->mirror_array[j]->efficiency) {
+				gmb();
 				swap(fls->mirror_array[i],
 				     fls->mirror_array[j]);
+			}
 	}
 }
 
@@ -452,7 +446,6 @@ ff_layout_alloc_lseg(struct pnfs_layout_hdr *lh,
 			goto out_err_free;
 
 		/* fh */
-		rc = -EIO;
 		p = xdr_inline_decode(&stream, 4);
 		if (!p)
 			goto out_err_free;
@@ -492,9 +485,9 @@ ff_layout_alloc_lseg(struct pnfs_layout_hdr *lh,
 		if (gfp_flags & __GFP_FS)
 			kcred = prepare_kernel_cred(NULL);
 		else {
-			unsigned int nofs_flags = memalloc_nofs_save();
+			unsigned int noio_flags = memalloc_noio_save();
 			kcred = prepare_kernel_cred(NULL);
-			memalloc_nofs_restore(nofs_flags);
+			memalloc_noio_restore(noio_flags);
 		}
 		rc = -ENOMEM;
 		if (!kcred)
@@ -625,11 +618,12 @@ nfs4_ff_layoutstat_start_io(struct nfs4_ff_layout_mirror *mirror,
 			    struct nfs4_ff_layoutstat *layoutstat,
 			    ktime_t now)
 {
+	static const ktime_t notime = {0};
 	s64 report_interval = FF_LAYOUTSTATS_REPORT_INTERVAL;
 	struct nfs4_flexfile_layout *ffl = FF_LAYOUT_FROM_HDR(mirror->layout);
 
 	nfs4_ff_start_busy_timer(&layoutstat->busy_timer, now);
-	if (!mirror->start_time)
+	if (ktime_equal(mirror->start_time, notime))
 		mirror->start_time = now;
 	if (mirror->report_interval != 0)
 		report_interval = (s64)mirror->report_interval * 1000LL;
@@ -852,7 +846,6 @@ ff_layout_pg_init_read(struct nfs_pageio_descriptor *pgio,
 	int ds_idx;
 
 retry:
-	pnfs_generic_pg_check_layout(pgio);
 	/* Use full layout for now */
 	if (!pgio->pg_lseg)
 		ff_layout_pg_get_read(pgio, req, false);
@@ -901,7 +894,6 @@ ff_layout_pg_init_write(struct nfs_pageio_descriptor *pgio,
 	int status;
 
 retry:
-	pnfs_generic_pg_check_layout(pgio);
 	if (!pgio->pg_lseg) {
 		pgio->pg_lseg = pnfs_update_layout(pgio->pg_inode,
 						   req->wb_context,
@@ -1055,10 +1047,37 @@ static int ff_layout_async_handle_error_v4(struct rpc_task *task,
 {
 	struct pnfs_layout_hdr *lo = lseg->pls_layout;
 	struct inode *inode = lo->plh_inode;
+	struct nfs_server *mds_server = NFS_SERVER(inode);
+
 	struct nfs4_deviceid_node *devid = FF_LAYOUT_DEVID_NODE(lseg, idx);
+	struct nfs_client *mds_client = mds_server->nfs_client;
 	struct nfs4_slot_table *tbl = &clp->cl_session->fc_slot_table;
 
+	if (task->tk_status >= 0)
+		return 0;
+
 	switch (task->tk_status) {
+	/* MDS state errors */
+	case -NFS4ERR_DELEG_REVOKED:
+	case -NFS4ERR_ADMIN_REVOKED:
+	case -NFS4ERR_BAD_STATEID:
+		if (state == NULL)
+			break;
+		nfs_remove_bad_delegation(state->inode, NULL);
+	case -NFS4ERR_OPENMODE:
+		if (state == NULL)
+			break;
+		if (nfs4_schedule_stateid_recovery(mds_server, state) < 0)
+			goto out_bad_stateid;
+		goto wait_on_recovery;
+	case -NFS4ERR_EXPIRED:
+		if (state != NULL) {
+			if (nfs4_schedule_stateid_recovery(mds_server, state) < 0)
+				goto out_bad_stateid;
+		}
+		nfs4_schedule_lease_recovery(mds_client);
+		goto wait_on_recovery;
+	/* DS session errors */
 	case -NFS4ERR_BADSESSION:
 	case -NFS4ERR_BADSLOT:
 	case -NFS4ERR_BAD_HIGH_SLOT:
@@ -1106,8 +1125,7 @@ static int ff_layout_async_handle_error_v4(struct rpc_task *task,
 	case -EPIPE:
 		dprintk("%s DS connection error %d\n", __func__,
 			task->tk_status);
-		nfs4_delete_deviceid(devid->ld, devid->nfs_client,
-				&devid->deviceid);
+		nfs4_mark_deviceid_unavailable(devid);
 		rpc_wake_up(&tbl->slot_tbl_waitq);
 		/* fall through */
 	default:
@@ -1118,8 +1136,17 @@ reset:
 			task->tk_status);
 		return -NFS4ERR_RESET_TO_MDS;
 	}
+out:
 	task->tk_status = 0;
 	return -EAGAIN;
+out_bad_stateid:
+	task->tk_status = -EIO;
+	return 0;
+wait_on_recovery:
+	rpc_sleep_on(&mds_client->cl_rpcwaitq, task, NULL);
+	if (test_bit(NFS4CLNT_MANAGER_RUNNING, &mds_client->cl_state) == 0)
+		rpc_wake_up_queued_task(&mds_client->cl_rpcwaitq, task);
+	goto out;
 }
 
 /* Retry all errors through either pNFS or MDS except for -EJUKEBOX */
@@ -1129,6 +1156,9 @@ static int ff_layout_async_handle_error_v3(struct rpc_task *task,
 {
 	struct nfs4_deviceid_node *devid = FF_LAYOUT_DEVID_NODE(lseg, idx);
 
+	if (task->tk_status >= 0)
+		return 0;
+
 	switch (task->tk_status) {
 	/* File access problems. Don't mark the device as unavailable */
 	case -EACCES:
@@ -1144,8 +1174,7 @@ static int ff_layout_async_handle_error_v3(struct rpc_task *task,
 	default:
 		dprintk("%s DS connection error %d\n", __func__,
 			task->tk_status);
-		nfs4_delete_deviceid(devid->ld, devid->nfs_client,
-				&devid->deviceid);
+		nfs4_mark_deviceid_unavailable(devid);
 	}
 	/* FIXME: Need to prevent infinite looping here. */
 	return -NFS4ERR_RESET_TO_PNFS;
@@ -1164,13 +1193,6 @@ static int ff_layout_async_handle_error(struct rpc_task *task,
 {
 	int vers = clp->cl_nfs_mod->rpc_vers->number;
 
-	if (task->tk_status >= 0)
-		return 0;
-
-	/* Handle the case of an invalid layout segment */
-	if (!pnfs_is_valid_lseg(lseg))
-		return -NFS4ERR_RESET_TO_PNFS;
-
 	switch (vers) {
 	case 3:
 		return ff_layout_async_handle_error_v3(task, lseg, idx);
@@ -1254,7 +1276,6 @@ static int ff_layout_read_done_cb(struct rpc_task *task,
 					hdr->pgio_mirror_idx + 1,
 					&hdr->pgio_mirror_idx))
 			goto out_eagain;
-		ff_layout_read_record_layoutstats_done(task, hdr);
 		pnfs_read_resend_pnfs(hdr);
 		return task->tk_status;
 	case -NFS4ERR_RESET_TO_MDS:
@@ -1360,14 +1381,30 @@ static void ff_layout_read_prepare_v3(struct rpc_task *task, void *data)
 	rpc_call_start(task);
 }
 
+static int ff_layout_setup_sequence(struct nfs_client *ds_clp,
+				    struct nfs4_sequence_args *args,
+				    struct nfs4_sequence_res *res,
+				    struct rpc_task *task)
+{
+	if (ds_clp->cl_session)
+		return nfs41_setup_sequence(ds_clp->cl_session,
+					   args,
+					   res,
+					   task);
+	return nfs40_setup_sequence(ds_clp->cl_slot_tbl,
+				   args,
+				   res,
+				   task);
+}
+
 static void ff_layout_read_prepare_v4(struct rpc_task *task, void *data)
 {
 	struct nfs_pgio_header *hdr = data;
 
-	if (nfs4_setup_sequence(hdr->ds_clp,
-				&hdr->args.seq_args,
-				&hdr->res.seq_res,
-				task))
+	if (ff_layout_setup_sequence(hdr->ds_clp,
+				     &hdr->args.seq_args,
+				     &hdr->res.seq_res,
+				     task))
 		return;
 
 	if (ff_layout_read_prepare_common(task, hdr))
@@ -1538,10 +1575,10 @@ static void ff_layout_write_prepare_v4(struct rpc_task *task, void *data)
 {
 	struct nfs_pgio_header *hdr = data;
 
-	if (nfs4_setup_sequence(hdr->ds_clp,
-				&hdr->args.seq_args,
-				&hdr->res.seq_res,
-				task))
+	if (ff_layout_setup_sequence(hdr->ds_clp,
+				     &hdr->args.seq_args,
+				     &hdr->res.seq_res,
+				     task))
 		return;
 
 	if (ff_layout_write_prepare_common(task, hdr))
@@ -1627,10 +1664,10 @@ static void ff_layout_commit_prepare_v4(struct rpc_task *task, void *data)
 {
 	struct nfs_commit_data *wdata = data;
 
-	if (nfs4_setup_sequence(wdata->ds_clp,
-				&wdata->args.seq_args,
-				&wdata->res.seq_res,
-				task))
+	if (ff_layout_setup_sequence(wdata->ds_clp,
+				 &wdata->args.seq_args,
+				 &wdata->res.seq_res,
+				 task))
 		return;
 	ff_layout_commit_prepare_common(task, data);
 }
@@ -1711,7 +1748,7 @@ ff_layout_read_pagelist(struct nfs_pgio_header *hdr)
 	int vers;
 	struct nfs_fh *fh;
 
-	dprintk("--> %s ino %lu pgbase %u req %zu@%llu\n",
+	dprintk("--> %s ino %lu pgbase %u req %Zu@%llu\n",
 		__func__, hdr->inode->i_ino,
 		hdr->args.pgbase, (size_t)hdr->args.count, offset);
 
@@ -1775,20 +1812,20 @@ ff_layout_write_pagelist(struct nfs_pgio_header *hdr, int sync)
 
 	ds = nfs4_ff_layout_prepare_ds(lseg, idx, true);
 	if (!ds)
-		goto out_failed;
+		return PNFS_NOT_ATTEMPTED;
 
 	ds_clnt = nfs4_ff_find_or_create_ds_client(lseg, idx, ds->ds_clp,
 						   hdr->inode);
 	if (IS_ERR(ds_clnt))
-		goto out_failed;
+		return PNFS_NOT_ATTEMPTED;
 
 	ds_cred = ff_layout_get_ds_cred(lseg, idx, hdr->cred);
 	if (!ds_cred)
-		goto out_failed;
+		return PNFS_NOT_ATTEMPTED;
 
 	vers = nfs4_ff_layout_ds_version(lseg, idx);
 
-	dprintk("%s ino %lu sync %d req %zu@%llu DS: %s cl_count %d vers %d\n",
+	dprintk("%s ino %lu sync %d req %Zu@%llu DS: %s cl_count %d vers %d\n",
 		__func__, hdr->inode->i_ino, sync, (size_t) hdr->args.count,
 		offset, ds->ds_remotestr, atomic_read(&ds->ds_clp->cl_count),
 		vers);
@@ -1814,11 +1851,6 @@ ff_layout_write_pagelist(struct nfs_pgio_header *hdr, int sync)
 			  sync, RPC_TASK_SOFTCONN);
 	put_cred(ds_cred);
 	return PNFS_ATTEMPTED;
-
-out_failed:
-	if (ff_layout_avoid_mds_available_ds(lseg))
-		return PNFS_TRY_AGAIN;
-	return PNFS_NOT_ATTEMPTED;
 }
 
 static u32 calc_ds_index_from_commit(struct pnfs_layout_segment *lseg, u32 i)
@@ -1847,10 +1879,6 @@ static int ff_layout_initiate_commit(struct nfs_commit_data *data, int how)
 	int vers, ret;
 	struct nfs_fh *fh;
 
-	if (!lseg || !(pnfs_is_valid_lseg(lseg) ||
-	    test_bit(NFS_LSEG_LAYOUTRETURN, &lseg->pls_flags)))
-		goto out_err;
-
 	idx = calc_ds_index_from_commit(lseg, data->ds_commit_index);
 	ds = nfs4_ff_layout_prepare_ds(lseg, idx, true);
 	if (!ds)
@@ -1916,85 +1944,38 @@ ff_layout_free_deviceid_node(struct nfs4_deviceid_node *d)
 						  id_node));
 }
 
-static int ff_layout_encode_ioerr(struct xdr_stream *xdr,
-				  const struct nfs4_layoutreturn_args *args,
-				  const struct nfs4_flexfile_layoutreturn_args *ff_args)
+static int ff_layout_encode_ioerr(struct nfs4_flexfile_layout *flo,
+				  struct xdr_stream *xdr,
+				  const struct nfs4_layoutreturn_args *args)
 {
+	struct pnfs_layout_hdr *hdr = &flo->generic_hdr;
 	__be32 *start;
+	int count = 0, ret = 0;
 
 	start = xdr_reserve_space(xdr, 4);
 	if (unlikely(!start))
 		return -E2BIG;
 
-	*start = cpu_to_be32(ff_args->num_errors);
 	/* This assume we always return _ALL_ layouts */
-	return ff_layout_encode_ds_ioerr(xdr, &ff_args->errors);
-}
+	spin_lock(&hdr->plh_inode->i_lock);
+	ret = ff_layout_encode_ds_ioerr(flo, xdr, &count, &args->range);
+	spin_unlock(&hdr->plh_inode->i_lock);
 
-static void
-encode_opaque_fixed(struct xdr_stream *xdr, const void *buf, size_t len)
-{
-	WARN_ON_ONCE(xdr_stream_encode_opaque_fixed(xdr, buf, len) < 0);
-}
+	*start = cpu_to_be32(count);
 
-static void
-ff_layout_encode_ff_iostat_head(struct xdr_stream *xdr,
-			    const nfs4_stateid *stateid,
-			    const struct nfs42_layoutstat_devinfo *devinfo)
-{
-	__be32 *p;
-
-	p = xdr_reserve_space(xdr, 8 + 8);
-	p = xdr_encode_hyper(p, devinfo->offset);
-	p = xdr_encode_hyper(p, devinfo->length);
-	encode_opaque_fixed(xdr, stateid->data, NFS4_STATEID_SIZE);
-	p = xdr_reserve_space(xdr, 4*8);
-	p = xdr_encode_hyper(p, devinfo->read_count);
-	p = xdr_encode_hyper(p, devinfo->read_bytes);
-	p = xdr_encode_hyper(p, devinfo->write_count);
-	p = xdr_encode_hyper(p, devinfo->write_bytes);
-	encode_opaque_fixed(xdr, devinfo->dev_id.data, NFS4_DEVICEID4_SIZE);
-}
-
-static void
-ff_layout_encode_ff_iostat(struct xdr_stream *xdr,
-			    const nfs4_stateid *stateid,
-			    const struct nfs42_layoutstat_devinfo *devinfo)
-{
-	ff_layout_encode_ff_iostat_head(xdr, stateid, devinfo);
-	ff_layout_encode_ff_layoutupdate(xdr, devinfo,
-			devinfo->ld_private.data);
+	return ret;
 }
 
 /* report nothing for now */
-static void ff_layout_encode_iostats_array(struct xdr_stream *xdr,
-		const struct nfs4_layoutreturn_args *args,
-		struct nfs4_flexfile_layoutreturn_args *ff_args)
+static void ff_layout_encode_iostats(struct nfs4_flexfile_layout *flo,
+				     struct xdr_stream *xdr,
+				     const struct nfs4_layoutreturn_args *args)
 {
 	__be32 *p;
-	int i;
 
 	p = xdr_reserve_space(xdr, 4);
-	*p = cpu_to_be32(ff_args->num_dev);
-	for (i = 0; i < ff_args->num_dev; i++)
-		ff_layout_encode_ff_iostat(xdr,
-				&args->layout->plh_stateid,
-				&ff_args->devinfo[i]);
-}
-
-static void
-ff_layout_free_iostats_array(struct nfs42_layoutstat_devinfo *devinfo,
-		unsigned int num_entries)
-{
-	unsigned int i;
-
-	for (i = 0; i < num_entries; i++) {
-		if (!devinfo[i].ld_private.ops)
-			continue;
-		if (!devinfo[i].ld_private.ops->free)
-			continue;
-		devinfo[i].ld_private.ops->free(&devinfo[i].ld_private);
-	}
+	if (likely(p))
+		*p = cpu_to_be32(0);
 }
 
 static struct nfs4_deviceid_node *
@@ -2010,91 +1991,24 @@ ff_layout_alloc_deviceid_node(struct nfs_server *server,
 }
 
 static void
-ff_layout_encode_layoutreturn(struct xdr_stream *xdr,
-		const void *voidargs,
-		const struct nfs4_xdr_opaque_data *ff_opaque)
+ff_layout_encode_layoutreturn(struct pnfs_layout_hdr *lo,
+			      struct xdr_stream *xdr,
+			      const struct nfs4_layoutreturn_args *args)
 {
-	const struct nfs4_layoutreturn_args *args = voidargs;
-	struct nfs4_flexfile_layoutreturn_args *ff_args = ff_opaque->data;
-	struct xdr_buf tmp_buf = {
-		.head = {
-			[0] = {
-				.iov_base = page_address(ff_args->pages[0]),
-			},
-		},
-		.buflen = PAGE_SIZE,
-	};
-	struct xdr_stream tmp_xdr;
+	struct nfs4_flexfile_layout *flo = FF_LAYOUT_FROM_HDR(lo);
 	__be32 *start;
 
 	dprintk("%s: Begin\n", __func__);
-
-	xdr_init_encode(&tmp_xdr, &tmp_buf, NULL, NULL);
-
-	ff_layout_encode_ioerr(&tmp_xdr, args, ff_args);
-	ff_layout_encode_iostats_array(&tmp_xdr, args, ff_args);
-
 	start = xdr_reserve_space(xdr, 4);
-	*start = cpu_to_be32(tmp_buf.len);
-	xdr_write_pages(xdr, ff_args->pages, 0, tmp_buf.len);
+	BUG_ON(!start);
 
+	ff_layout_encode_ioerr(flo, xdr, args);
+	ff_layout_encode_iostats(flo, xdr, args);
+
+	*start = cpu_to_be32((xdr->p - start - 1) * 4);
 	dprintk("%s: Return\n", __func__);
 }
 
-static void
-ff_layout_free_layoutreturn(struct nfs4_xdr_opaque_data *args)
-{
-	struct nfs4_flexfile_layoutreturn_args *ff_args;
-
-	if (!args->data)
-		return;
-	ff_args = args->data;
-	args->data = NULL;
-
-	ff_layout_free_ds_ioerr(&ff_args->errors);
-	ff_layout_free_iostats_array(ff_args->devinfo, ff_args->num_dev);
-
-	put_page(ff_args->pages[0]);
-	kfree(ff_args);
-}
-
-static const struct nfs4_xdr_opaque_ops layoutreturn_ops = {
-	.encode = ff_layout_encode_layoutreturn,
-	.free = ff_layout_free_layoutreturn,
-};
-
-static int
-ff_layout_prepare_layoutreturn(struct nfs4_layoutreturn_args *args)
-{
-	struct nfs4_flexfile_layoutreturn_args *ff_args;
-	struct nfs4_flexfile_layout *ff_layout = FF_LAYOUT_FROM_HDR(args->layout);
-
-	ff_args = kmalloc(sizeof(*ff_args), GFP_KERNEL);
-	if (!ff_args)
-		goto out_nomem;
-	ff_args->pages[0] = alloc_page(GFP_KERNEL);
-	if (!ff_args->pages[0])
-		goto out_nomem_free;
-
-	INIT_LIST_HEAD(&ff_args->errors);
-	ff_args->num_errors = ff_layout_fetch_ds_ioerr(args->layout,
-			&args->range, &ff_args->errors,
-			FF_LAYOUTRETURN_MAXERR);
-
-	spin_lock(&args->inode->i_lock);
-	ff_args->num_dev = ff_layout_mirror_prepare_stats(&ff_layout->generic_hdr,
-			&ff_args->devinfo[0], ARRAY_SIZE(ff_args->devinfo));
-	spin_unlock(&args->inode->i_lock);
-
-	args->ld_private->ops = &layoutreturn_ops;
-	args->ld_private->data = ff_args;
-	return 0;
-out_nomem_free:
-	kfree(ff_args);
-out_nomem:
-	return -ENOMEM;
-}
-
 static int
 ff_layout_ntop4(const struct sockaddr *sap, char *buf, const size_t buflen)
 {
@@ -2215,18 +2129,21 @@ ff_layout_encode_io_latency(struct xdr_stream *xdr,
 }
 
 static void
-ff_layout_encode_ff_layoutupdate(struct xdr_stream *xdr,
-			      const struct nfs42_layoutstat_devinfo *devinfo,
-			      struct nfs4_ff_layout_mirror *mirror)
+ff_layout_encode_layoutstats(struct xdr_stream *xdr,
+			     const struct nfs42_layoutstat_args *args,
+			     struct nfs42_layoutstat_devinfo *devinfo)
 {
+	struct nfs4_ff_layout_mirror *mirror = devinfo->layout_private;
 	struct nfs4_pnfs_ds_addr *da;
 	struct nfs4_pnfs_ds *ds = mirror->mirror_ds->ds;
 	struct nfs_fh *fh = &mirror->fh_versions[0];
-	__be32 *p;
+	__be32 *p, *start;
 
 	da = list_first_entry(&ds->ds_addrs, struct nfs4_pnfs_ds_addr, da_node);
 	dprintk("%s: DS %s: encoding address %s\n",
 		__func__, ds->ds_remotestr, da->da_remotestr);
+	/* layoutupdate length */
+	start = xdr_reserve_space(xdr, 4);
 	/* netaddr4 */
 	ff_layout_encode_netaddr(xdr, da);
 	/* nfs_fh4 */
@@ -2243,44 +2160,19 @@ ff_layout_encode_ff_layoutupdate(struct xdr_stream *xdr,
 	/* bool */
 	p = xdr_reserve_space(xdr, 4);
 	*p = cpu_to_be32(false);
-}
-
-static void
-ff_layout_encode_layoutstats(struct xdr_stream *xdr, const void *args,
-			     const struct nfs4_xdr_opaque_data *opaque)
-{
-	struct nfs42_layoutstat_devinfo *devinfo = container_of(opaque,
-			struct nfs42_layoutstat_devinfo, ld_private);
-	__be32 *start;
-
-	/* layoutupdate length */
-	start = xdr_reserve_space(xdr, 4);
-	ff_layout_encode_ff_layoutupdate(xdr, devinfo, opaque->data);
 
 	*start = cpu_to_be32((xdr->p - start - 1) * 4);
 }
 
-static void
-ff_layout_free_layoutstats(struct nfs4_xdr_opaque_data *opaque)
-{
-	struct nfs4_ff_layout_mirror *mirror = opaque->data;
-
-	ff_layout_put_mirror(mirror);
-}
-
-static const struct nfs4_xdr_opaque_ops layoutstat_ops = {
-	.encode = ff_layout_encode_layoutstats,
-	.free	= ff_layout_free_layoutstats,
-};
-
 static int
-ff_layout_mirror_prepare_stats(struct pnfs_layout_hdr *lo,
-			       struct nfs42_layoutstat_devinfo *devinfo,
+ff_layout_mirror_prepare_stats(struct nfs42_layoutstat_args *args,
+			       struct pnfs_layout_hdr *lo,
 			       int dev_limit)
 {
 	struct nfs4_flexfile_layout *ff_layout = FF_LAYOUT_FROM_HDR(lo);
 	struct nfs4_ff_layout_mirror *mirror;
 	struct nfs4_deviceid_node *dev;
+	struct nfs42_layoutstat_devinfo *devinfo;
 	int i = 0;
 
 	list_for_each_entry(mirror, &ff_layout->mirrors, mirrors) {
@@ -2294,20 +2186,18 @@ ff_layout_mirror_prepare_stats(struct pnfs_layout_hdr *lo,
 		if (!atomic_inc_not_zero(&mirror->ref))
 			continue;
 		dev = &mirror->mirror_ds->id_node; 
+		devinfo = &args->devinfo[i];
 		memcpy(&devinfo->dev_id, &dev->deviceid, NFS4_DEVICEID4_SIZE);
 		devinfo->offset = 0;
 		devinfo->length = NFS4_MAX_UINT64;
-		spin_lock(&mirror->lock);
 		devinfo->read_count = mirror->read_stat.io_stat.ops_completed;
 		devinfo->read_bytes = mirror->read_stat.io_stat.bytes_completed;
 		devinfo->write_count = mirror->write_stat.io_stat.ops_completed;
 		devinfo->write_bytes = mirror->write_stat.io_stat.bytes_completed;
-		spin_unlock(&mirror->lock);
 		devinfo->layout_type = LAYOUT_FLEX_FILES;
-		devinfo->ld_private.ops = &layoutstat_ops;
-		devinfo->ld_private.data = mirror;
+		devinfo->layoutstats_encode = ff_layout_encode_layoutstats;
+		devinfo->layout_private = mirror;
 
-		devinfo++;
 		i++;
 	}
 	return i;
@@ -2317,17 +2207,29 @@ static int
 ff_layout_prepare_layoutstats(struct nfs42_layoutstat_args *args)
 {
 	struct nfs4_flexfile_layout *ff_layout;
-	const int dev_count = PNFS_LAYOUTSTATS_MAXDEV;
+	struct nfs4_ff_layout_mirror *mirror;
+	int dev_count = 0;
 
+	spin_lock(&args->inode->i_lock);
+	ff_layout = FF_LAYOUT_FROM_HDR(NFS_I(args->inode)->layout);
+	list_for_each_entry(mirror, &ff_layout->mirrors, mirrors) {
+		if (atomic_read(&mirror->ref) != 0)
+			dev_count ++;
+	}
+	spin_unlock(&args->inode->i_lock);
 	/* For now, send at most PNFS_LAYOUTSTATS_MAXDEV statistics */
+	if (dev_count > PNFS_LAYOUTSTATS_MAXDEV) {
+		dprintk("%s: truncating devinfo to limit (%d:%d)\n",
+			__func__, dev_count, PNFS_LAYOUTSTATS_MAXDEV);
+		dev_count = PNFS_LAYOUTSTATS_MAXDEV;
+	}
 	args->devinfo = kmalloc_array(dev_count, sizeof(*args->devinfo), GFP_NOIO);
 	if (!args->devinfo)
 		return -ENOMEM;
 
 	spin_lock(&args->inode->i_lock);
-	ff_layout = FF_LAYOUT_FROM_HDR(NFS_I(args->inode)->layout);
-	args->num_dev = ff_layout_mirror_prepare_stats(&ff_layout->generic_hdr,
-			&args->devinfo[0], dev_count);
+	args->num_dev = ff_layout_mirror_prepare_stats(args,
+			&ff_layout->generic_hdr, dev_count);
 	spin_unlock(&args->inode->i_lock);
 	if (!args->num_dev) {
 		kfree(args->devinfo);
@@ -2338,21 +2240,23 @@ ff_layout_prepare_layoutstats(struct nfs42_layoutstat_args *args)
 	return 0;
 }
 
-static int
-ff_layout_set_layoutdriver(struct nfs_server *server,
-		const struct nfs_fh *dummy)
+static void
+ff_layout_cleanup_layoutstats(struct nfs42_layoutstat_data *data)
 {
-#if IS_ENABLED(CONFIG_NFS_V4_2)
-	server->caps |= NFS_CAP_LAYOUTSTATS;
-#endif
-	return 0;
+	struct nfs4_ff_layout_mirror *mirror;
+	int i;
+
+	for (i = 0; i < data->args.num_dev; i++) {
+		mirror = data->args.devinfo[i].layout_private;
+		data->args.devinfo[i].layout_private = NULL;
+		ff_layout_put_mirror(mirror);
+	}
 }
 
 static struct pnfs_layoutdriver_type flexfilelayout_type = {
 	.id			= LAYOUT_FLEX_FILES,
 	.name			= "LAYOUT_FLEX_FILES",
 	.owner			= THIS_MODULE,
-	.set_layoutdriver	= ff_layout_set_layoutdriver,
 	.alloc_layout_hdr	= ff_layout_alloc_layout_hdr,
 	.free_layout_hdr	= ff_layout_free_layout_hdr,
 	.alloc_lseg		= ff_layout_alloc_lseg,
@@ -2370,9 +2274,10 @@ static struct pnfs_layoutdriver_type flexfilelayout_type = {
 	.read_pagelist		= ff_layout_read_pagelist,
 	.write_pagelist		= ff_layout_write_pagelist,
 	.alloc_deviceid_node    = ff_layout_alloc_deviceid_node,
-	.prepare_layoutreturn   = ff_layout_prepare_layoutreturn,
+	.encode_layoutreturn    = ff_layout_encode_layoutreturn,
 	.sync			= pnfs_nfs_generic_sync,
 	.prepare_layoutstats	= ff_layout_prepare_layoutstats,
+	.cleanup_layoutstats	= ff_layout_cleanup_layoutstats,
 };
 
 static int __init nfs4flexfilelayout_init(void)
@@ -2381,6 +2286,11 @@ static int __init nfs4flexfilelayout_init(void)
 
 	printk(KERN_INFO "%s: NFSv4 Flexfile Layout Driver Registering...\n",
 	       __func__);
+	if (!ff_zero_group) {
+		ff_zero_group = groups_alloc(0);
+		if (!ff_zero_group)
+			return -ENOMEM;
+	}
 	return pnfs_register_layoutdriver(&flexfilelayout_type);
 }
 
@@ -2389,6 +2299,10 @@ static void __exit nfs4flexfilelayout_exit(void)
 	printk(KERN_INFO "%s: NFSv4 Flexfile Layout Driver Unregistering...\n",
 	       __func__);
 	pnfs_unregister_layoutdriver(&flexfilelayout_type);
+	if (ff_zero_group) {
+		put_group_info(ff_zero_group);
+		ff_zero_group = NULL;
+	}
 }
 
 MODULE_ALIAS("nfs-layouttype4-4");
