@@ -45,6 +45,7 @@
 #include <linux/key-type.h>
 #include <keys/user-type.h>
 #include <linux/module.h>
+#include <linux/nospec.h>
 #include <linux/old-kernel.h>
 
 #include "internal.h"
@@ -176,9 +177,8 @@ static int nfs_map_numeric_to_string(__u32 id, char *buf, size_t buflen)
 
 static struct key_type key_type_id_resolver = {
 	.name		= "id_resolver",
-	.preparse	= user_preparse,
-	.free_preparse	= user_free_preparse,
-	.instantiate	= generic_key_instantiate,
+	.instantiate	= user_instantiate,
+	.match		= user_match,
 	.revoke		= user_revoke,
 	.destroy	= user_destroy,
 	.describe	= user_describe,
@@ -202,7 +202,7 @@ int nfs_idmap_init(void)
 				GLOBAL_ROOT_UID, GLOBAL_ROOT_GID, cred,
 				(KEY_POS_ALL & ~KEY_POS_SETATTR) |
 				KEY_USR_VIEW | KEY_USR_READ,
-				KEY_ALLOC_NOT_IN_QUOTA, NULL, NULL);
+				KEY_ALLOC_NOT_IN_QUOTA, NULL);
 	if (IS_ERR(keyring)) {
 		ret = PTR_ERR(keyring);
 		goto failed_put_cred;
@@ -275,7 +275,7 @@ static struct key *nfs_idmap_request_key(const char *name, size_t namelen,
 	ssize_t ret;
 
 	ret = nfs_idmap_get_desc(name, namelen, type, strlen(type), &desc);
-	if (ret < 0)
+	if (ret <= 0)
 		return ERR_PTR(ret);
 
 	rkey = request_key(&key_type_id_resolver, desc, "");
@@ -344,7 +344,7 @@ static ssize_t nfs_idmap_lookup_name(__u32 id, const char *type, char *buf,
 	int id_len;
 	ssize_t ret;
 
-	id_len = nfs_map_numeric_to_string(id, id_str, sizeof(id_str));
+	id_len = snprintf(id_str, sizeof(id_str), "%u", id);
 	ret = nfs_idmap_get_key(id_str, id_len, type, buf, buflen, idmap);
 	if (ret < 0)
 		return -EINVAL;
@@ -365,8 +365,7 @@ static int nfs_idmap_lookup_id(const char *name, size_t namelen, const char *typ
 		ret = -EINVAL;
 	} else {
 		ret = kstrtol(id_str, 10, &id_long);
-		if (!ret)
-			*id = (__u32)id_long;
+		*id = (__u32)id_long;
 	}
 	return ret;
 }
@@ -400,9 +399,8 @@ static const struct rpc_pipe_ops idmap_upcall_ops = {
 
 static struct key_type key_type_id_resolver_legacy = {
 	.name		= "id_legacy",
-	.preparse	= user_preparse,
-	.free_preparse	= user_free_preparse,
-	.instantiate	= generic_key_instantiate,
+	.instantiate	= user_instantiate,
+	.match		= user_match,
 	.revoke		= user_revoke,
 	.destroy	= user_destroy,
 	.describe	= user_describe,
@@ -624,14 +622,13 @@ static int nfs_idmap_read_and_verify_message(struct idmap_msg *im,
 		if (strcmp(upcall->im_name, im->im_name) != 0)
 			break;
 		/* Note: here we store the NUL terminator too */
-		len = 1 + nfs_map_numeric_to_string(im->im_id, id_str,
-						    sizeof(id_str));
+		len = sprintf(id_str, "%d", im->im_id) + 1;
 		ret = nfs_idmap_instantiate(key, authkey, id_str, len);
 		break;
 	case IDMAP_CONV_IDTONAME:
 		if (upcall->im_id != im->im_id)
 			break;
-		len = strlen(im->im_name);
+		len = array_index_nospec(strlen(im->im_name), IDMAP_NAMESZ);
 		ret = nfs_idmap_instantiate(key, authkey, im->im_name, len);
 		break;
 	default:
