@@ -1112,80 +1112,6 @@ static void cache_dequeue(struct cache_detail *detail, struct cache_head *ch)
 	}
 }
 
-#define ESCAPE_OCTAL            0x08
-
-static bool escape_passthrough(unsigned char c, char **dst, char *end)
-{
-	char *out = *dst;
-
-	if (out < end)
-		*out = c;
-	*dst = out + 1;
-	return true;
-}
-
-static bool escape_octal(unsigned char c, char **dst, char *end)
-{
-	char *out = *dst;
-
-	if (out < end)
-		*out = '\\';
-	++out;
-	if (out < end)
-		*out = ((c >> 6) & 0x07) + '0';
-	++out;
-	if (out < end)
-		*out = ((c >> 3) & 0x07) + '0';
-	++out;
-	if (out < end)
-		*out = ((c >> 0) & 0x07) + '0';
-	++out;
-
-	*dst = out;
-	return true;
-}
-
-static int string_escape_mem(const char *src, size_t isz, char *dst, size_t osz,
-			     unsigned int flags, const char *only)
-{
-	char *p = dst;
-	char *end = p + osz;
-	bool is_dict = only && *only;
-
-	while (isz--) {
-		unsigned char c = *src++;
-
-		/*
-		 * Apply rules in the following sequence:
-		 *	- the character is printable, when @flags has
-		 *	  %ESCAPE_NP bit set
-		 *	- the @only string is supplied and does not contain a
-		 *	  character under question
-		 *	- the character doesn't fall into a class of symbols
-		 *	  defined by given @flags
-		 * In these cases we just pass through a character to the
-		 * output buffer.
-		 */
-		if ((is_dict && !strchr(only, c))) {
-			/* do nothing */
-		} else {
-			/* ESCAPE_OCTAL and ESCAPE_HEX always go last */
-			if (flags & ESCAPE_OCTAL && escape_octal(c, &p, end))
-				continue;
-		}
-
-		escape_passthrough(c, &p, end);
-	}
-
-	return p - dst;
-}
-
-static inline int string_escape_str(const char *src, char *dst, size_t sz,
-                unsigned int flags, const char *only)
-{
-        return string_escape_mem(src, strlen(src), dst, sz, flags, only);
-}
-
 /*
  * Support routines for text-based upcalls.
  * Fields are separated by spaces.
@@ -1503,7 +1429,7 @@ static int c_show(struct seq_file *m, void *p)
 		return cd->cache_show(m, cd, NULL);
 
 	ifdebug(CACHE)
-		seq_printf(m, "# expiry=%ld refcnt=%d flags=%lx\n",
+		seq_printf(m, "# expiry=%lld refcnt=%d flags=%lx\n",
 			   convert_to_wallclock(cp->expiry_time),
 			   kref_read(&cp->ref), cp->flags);
 	cache_get(cp);
@@ -1576,7 +1502,7 @@ static ssize_t read_flush(struct file *file, char __user *buf,
 	char tbuf[22];
 	size_t len;
 
-	len = snprintf(tbuf, sizeof(tbuf), "%lu\n",
+	len = snprintf(tbuf, sizeof(tbuf), "%llu\n",
 			convert_to_wallclock(cd->flush_time));
 	return simple_read_from_buffer(buf, count, ppos, tbuf, len);
 }
